# Copyright Amazon.com, Inc. or its affiliates. All rights reserved.
# SPDX-License-Identifier: Apache-2.0.

# This file is generated

import awsiot
import concurrent.futures
import datetime
import typing

class IotJobsClient(awsiot.MqttServiceClient):
    """

    The AWS IoT jobs service can be used to define a set of remote operations that are sent to and executed on one or more devices connected to AWS IoT.

    AWS Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#jobs-mqtt-api

    """

    def publish_describe_job_execution(self, request, qos):
        # type: (DescribeJobExecutionRequest, int) -> concurrent.futures.Future
        """

        Gets detailed information about a job execution.

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-describejobexecution

        Args:
            request: `DescribeJobExecutionRequest` instance.
            qos: The Quality of Service guarantee of this message

        Returns:
            A Future whose result will be None if the
            request is successfully published. The Future's result will be an
            exception if the request cannot be published.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")
        if not request.job_id:
            raise ValueError("request.job_id is required")

        return self._publish_operation(
            topic='$aws/things/{0.thing_name}/jobs/{0.job_id}/get'.format(request),
            qos=qos,
            payload=request.to_payload())

    def publish_get_pending_job_executions(self, request, qos):
        # type: (GetPendingJobExecutionsRequest, int) -> concurrent.futures.Future
        """

        Gets the list of all jobs for a thing that are not in a terminal state.

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-getpendingjobexecutions

        Args:
            request: `GetPendingJobExecutionsRequest` instance.
            qos: The Quality of Service guarantee of this message

        Returns:
            A Future whose result will be None if the
            request is successfully published. The Future's result will be an
            exception if the request cannot be published.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        return self._publish_operation(
            topic='$aws/things/{0.thing_name}/jobs/get'.format(request),
            qos=qos,
            payload=request.to_payload())

    def publish_start_next_pending_job_execution(self, request, qos):
        # type: (StartNextPendingJobExecutionRequest, int) -> concurrent.futures.Future
        """

        Gets and starts the next pending job execution for a thing (status IN_PROGRESS or QUEUED).

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-startnextpendingjobexecution

        Args:
            request: `StartNextPendingJobExecutionRequest` instance.
            qos: The Quality of Service guarantee of this message

        Returns:
            A Future whose result will be None if the
            request is successfully published. The Future's result will be an
            exception if the request cannot be published.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        return self._publish_operation(
            topic='$aws/things/{0.thing_name}/jobs/start-next'.format(request),
            qos=qos,
            payload=request.to_payload())

    def publish_update_job_execution(self, request, qos):
        # type: (UpdateJobExecutionRequest, int) -> concurrent.futures.Future
        """

        Updates the status of a job execution. You can optionally create a step timer by setting a value for the stepTimeoutInMinutes property. If you don't update the value of this property by running UpdateJobExecution again, the job execution times out when the step timer expires.

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-updatejobexecution

        Args:
            request: `UpdateJobExecutionRequest` instance.
            qos: The Quality of Service guarantee of this message

        Returns:
            A Future whose result will be None if the
            request is successfully published. The Future's result will be an
            exception if the request cannot be published.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")
        if not request.job_id:
            raise ValueError("request.job_id is required")

        return self._publish_operation(
            topic='$aws/things/{0.thing_name}/jobs/{0.job_id}/update'.format(request),
            qos=qos,
            payload=request.to_payload())

    def subscribe_to_describe_job_execution_accepted(self, request, qos, callback):
        # type: (DescribeJobExecutionSubscriptionRequest, int, typing.Callable[[DescribeJobExecutionResponse], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to the accepted topic for the DescribeJobExecution operation

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-describejobexecution

        Args:
            request: `DescribeJobExecutionSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `DescribeJobExecutionResponse`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")
        if not request.job_id:
            raise ValueError("request.job_id is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/{0.job_id}/get/accepted'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=DescribeJobExecutionResponse.from_payload)

    def subscribe_to_describe_job_execution_rejected(self, request, qos, callback):
        # type: (DescribeJobExecutionSubscriptionRequest, int, typing.Callable[[RejectedError], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to the rejected topic for the DescribeJobExecution operation

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-describejobexecution

        Args:
            request: `DescribeJobExecutionSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `RejectedError`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")
        if not request.job_id:
            raise ValueError("request.job_id is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/{0.job_id}/get/rejected'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=RejectedError.from_payload)

    def subscribe_to_get_pending_job_executions_accepted(self, request, qos, callback):
        # type: (GetPendingJobExecutionsSubscriptionRequest, int, typing.Callable[[GetPendingJobExecutionsResponse], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to the accepted topic for the GetPendingJobsExecutions operation

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-getpendingjobexecutions

        Args:
            request: `GetPendingJobExecutionsSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `GetPendingJobExecutionsResponse`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/get/accepted'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=GetPendingJobExecutionsResponse.from_payload)

    def subscribe_to_get_pending_job_executions_rejected(self, request, qos, callback):
        # type: (GetPendingJobExecutionsSubscriptionRequest, int, typing.Callable[[RejectedError], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to the rejected topic for the GetPendingJobsExecutions operation

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-getpendingjobexecutions

        Args:
            request: `GetPendingJobExecutionsSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `RejectedError`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/get/rejected'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=RejectedError.from_payload)

    def subscribe_to_job_executions_changed_events(self, request, qos, callback):
        # type: (JobExecutionsChangedSubscriptionRequest, int, typing.Callable[[JobExecutionsChangedEvent], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to JobExecutionsChanged notifications for a given IoT thing.

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-jobexecutionschanged

        Args:
            request: `JobExecutionsChangedSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `JobExecutionsChangedEvent`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/notify'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=JobExecutionsChangedEvent.from_payload)

    def subscribe_to_next_job_execution_changed_events(self, request, qos, callback):
        # type: (NextJobExecutionChangedSubscriptionRequest, int, typing.Callable[[NextJobExecutionChangedEvent], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-nextjobexecutionchanged

        Args:
            request: `NextJobExecutionChangedSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `NextJobExecutionChangedEvent`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/notify-next'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=NextJobExecutionChangedEvent.from_payload)

    def subscribe_to_start_next_pending_job_execution_accepted(self, request, qos, callback):
        # type: (StartNextPendingJobExecutionSubscriptionRequest, int, typing.Callable[[StartNextJobExecutionResponse], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to the accepted topic for the StartNextPendingJobExecution operation

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-startnextpendingjobexecution

        Args:
            request: `StartNextPendingJobExecutionSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `StartNextJobExecutionResponse`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/start-next/accepted'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=StartNextJobExecutionResponse.from_payload)

    def subscribe_to_start_next_pending_job_execution_rejected(self, request, qos, callback):
        # type: (StartNextPendingJobExecutionSubscriptionRequest, int, typing.Callable[[RejectedError], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to the rejected topic for the StartNextPendingJobExecution operation

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-startnextpendingjobexecution

        Args:
            request: `StartNextPendingJobExecutionSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `RejectedError`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/start-next/rejected'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=RejectedError.from_payload)

    def subscribe_to_update_job_execution_accepted(self, request, qos, callback):
        # type: (UpdateJobExecutionSubscriptionRequest, int, typing.Callable[[UpdateJobExecutionResponse], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to the accepted topic for the UpdateJobExecution operation

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-updatejobexecution

        Args:
            request: `UpdateJobExecutionSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `UpdateJobExecutionResponse`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.job_id:
            raise ValueError("request.job_id is required")
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/{0.job_id}/update/accepted'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=UpdateJobExecutionResponse.from_payload)

    def subscribe_to_update_job_execution_rejected(self, request, qos, callback):
        # type: (UpdateJobExecutionSubscriptionRequest, int, typing.Callable[[RejectedError], None]) -> typing.Tuple[concurrent.futures.Future, str]
        """

        Subscribes to the rejected topic for the UpdateJobExecution operation

        API Docs: https://docs.aws.amazon.com/iot/latest/developerguide/jobs-api.html#mqtt-updatejobexecution

        Args:
            request: `UpdateJobExecutionSubscriptionRequest` instance.
            qos: The Quality of Service guarantee of this message
            callback: Callback to invoke each time the event is received.
                The callback should take 1 argument of type `RejectedError`.
                The callback is not expected to return anything.

        Returns:
            Tuple with two values. The first is a `Future` whose result will be the
            `awscrt.mqtt.QoS` granted by the server, or an exception if the
            subscription fails. The second value is a topic which may be passed
            to `unsubscribe()` to stop receiving messages. Note that messages
            may arrive before the subscription is acknowledged.
        """
        if not request.job_id:
            raise ValueError("request.job_id is required")
        if not request.thing_name:
            raise ValueError("request.thing_name is required")

        if not callable(callback):
            raise ValueError("callback is required")

        return self._subscribe_operation(
            topic='$aws/things/{0.thing_name}/jobs/{0.job_id}/update/rejected'.format(request),
            qos=qos,
            callback=callback,
            payload_to_class_fn=RejectedError.from_payload)

class DescribeJobExecutionRequest(awsiot.ModeledClass):
    """

    Data needed to make a DescribeJobExecution request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): An opaque string used to correlate requests and responses. Enter an arbitrary value here and it is reflected in the response.
        execution_number (int): Optional. A number that identifies a job execution on a device. If not specified, the latest job execution is returned.
        include_job_document (bool): Optional. Unless set to false, the response contains the job document. The default is true.
        job_id (str): The unique identifier assigned to this job when it was created. Or use $next to return the next pending job execution for a thing (status IN_PROGRESS or QUEUED). In this case, any job executions with status IN_PROGRESS are returned first. Job executions are returned in the order in which they were created.
        thing_name (str): The name of the thing associated with the device.

    Attributes:
        client_token (str): An opaque string used to correlate requests and responses. Enter an arbitrary value here and it is reflected in the response.
        execution_number (int): Optional. A number that identifies a job execution on a device. If not specified, the latest job execution is returned.
        include_job_document (bool): Optional. Unless set to false, the response contains the job document. The default is true.
        job_id (str): The unique identifier assigned to this job when it was created. Or use $next to return the next pending job execution for a thing (status IN_PROGRESS or QUEUED). In this case, any job executions with status IN_PROGRESS are returned first. Job executions are returned in the order in which they were created.
        thing_name (str): The name of the thing associated with the device.
    """

    __slots__ = ['client_token', 'execution_number', 'include_job_document', 'job_id', 'thing_name']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.execution_number = kwargs.get('execution_number')
        self.include_job_document = kwargs.get('include_job_document')
        self.job_id = kwargs.get('job_id')
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'execution_number', 'include_job_document', 'job_id', 'thing_name'], args):
            setattr(self, key, val)

    def to_payload(self):
        # type: () -> typing.Dict[str, typing.Any]
        payload = {} # type: typing.Dict[str, typing.Any]
        if self.client_token is not None:
            payload['clientToken'] = self.client_token
        if self.execution_number is not None:
            payload['executionNumber'] = self.execution_number
        if self.include_job_document is not None:
            payload['includeJobDocument'] = self.include_job_document
        return payload

class DescribeJobExecutionResponse(awsiot.ModeledClass):
    """

    Response payload to a DescribeJobExecution request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): A client token used to correlate requests and responses.
        execution (JobExecutionData): Contains data about a job execution.
        timestamp (datetime.datetime): The time when the message was sent.

    Attributes:
        client_token (str): A client token used to correlate requests and responses.
        execution (JobExecutionData): Contains data about a job execution.
        timestamp (datetime.datetime): The time when the message was sent.
    """

    __slots__ = ['client_token', 'execution', 'timestamp']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.execution = kwargs.get('execution')
        self.timestamp = kwargs.get('timestamp')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'execution', 'timestamp'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> DescribeJobExecutionResponse
        new = cls()
        val = payload.get('clientToken')
        if val is not None:
            new.client_token = val
        val = payload.get('execution')
        if val is not None:
            new.execution = JobExecutionData.from_payload(val)
        val = payload.get('timestamp')
        if val is not None:
            new.timestamp = datetime.datetime.fromtimestamp(val)
        return new

class DescribeJobExecutionSubscriptionRequest(awsiot.ModeledClass):
    """

    Data needed to subscribe to DescribeJobExecution responses.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        job_id (str): Job ID that you want to subscribe to DescribeJobExecution response events for.
        thing_name (str): Name of the IoT Thing that you want to subscribe to DescribeJobExecution response events for.

    Attributes:
        job_id (str): Job ID that you want to subscribe to DescribeJobExecution response events for.
        thing_name (str): Name of the IoT Thing that you want to subscribe to DescribeJobExecution response events for.
    """

    __slots__ = ['job_id', 'thing_name']

    def __init__(self, *args, **kwargs):
        self.job_id = kwargs.get('job_id')
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['job_id', 'thing_name'], args):
            setattr(self, key, val)

class GetPendingJobExecutionsRequest(awsiot.ModeledClass):
    """

    Data needed to make a GetPendingJobExecutions request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): Optional. A client token used to correlate requests and responses. Enter an arbitrary value here and it is reflected in the response.
        thing_name (str): IoT Thing the request is relative to.

    Attributes:
        client_token (str): Optional. A client token used to correlate requests and responses. Enter an arbitrary value here and it is reflected in the response.
        thing_name (str): IoT Thing the request is relative to.
    """

    __slots__ = ['client_token', 'thing_name']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'thing_name'], args):
            setattr(self, key, val)

    def to_payload(self):
        # type: () -> typing.Dict[str, typing.Any]
        payload = {} # type: typing.Dict[str, typing.Any]
        if self.client_token is not None:
            payload['clientToken'] = self.client_token
        return payload

class GetPendingJobExecutionsResponse(awsiot.ModeledClass):
    """

    Response payload to a GetPendingJobExecutions request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): A client token used to correlate requests and responses.
        in_progress_jobs (typing.List[JobExecutionSummary]): A list of JobExecutionSummary objects with status IN_PROGRESS.
        queued_jobs (typing.List[JobExecutionSummary]): A list of JobExecutionSummary objects with status QUEUED.
        timestamp (datetime.datetime): The time when the message was sent.

    Attributes:
        client_token (str): A client token used to correlate requests and responses.
        in_progress_jobs (typing.List[JobExecutionSummary]): A list of JobExecutionSummary objects with status IN_PROGRESS.
        queued_jobs (typing.List[JobExecutionSummary]): A list of JobExecutionSummary objects with status QUEUED.
        timestamp (datetime.datetime): The time when the message was sent.
    """

    __slots__ = ['client_token', 'in_progress_jobs', 'queued_jobs', 'timestamp']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.in_progress_jobs = kwargs.get('in_progress_jobs')
        self.queued_jobs = kwargs.get('queued_jobs')
        self.timestamp = kwargs.get('timestamp')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'in_progress_jobs', 'queued_jobs', 'timestamp'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> GetPendingJobExecutionsResponse
        new = cls()
        val = payload.get('clientToken')
        if val is not None:
            new.client_token = val
        val = payload.get('inProgressJobs')
        if val is not None:
            new.in_progress_jobs = [JobExecutionSummary.from_payload(i) for i in val]
        val = payload.get('queuedJobs')
        if val is not None:
            new.queued_jobs = [JobExecutionSummary.from_payload(i) for i in val]
        val = payload.get('timestamp')
        if val is not None:
            new.timestamp = datetime.datetime.fromtimestamp(val)
        return new

class GetPendingJobExecutionsSubscriptionRequest(awsiot.ModeledClass):
    """

    Data needed to subscribe to GetPendingJobExecutions responses.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        thing_name (str): Name of the IoT Thing that you want to subscribe to GetPendingJobExecutions response events for.

    Attributes:
        thing_name (str): Name of the IoT Thing that you want to subscribe to GetPendingJobExecutions response events for.
    """

    __slots__ = ['thing_name']

    def __init__(self, *args, **kwargs):
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['thing_name'], args):
            setattr(self, key, val)

class JobExecutionData(awsiot.ModeledClass):
    """

    Data about a job execution.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        execution_number (int): A number that identifies a job execution on a device. It can be used later in commands that return or update job execution information.
        job_document (typing.Dict[str, typing.Any]): The content of the job document.
        job_id (str): The unique identifier you assigned to this job when it was created.
        last_updated_at (datetime.datetime): The time when the job execution started.
        queued_at (datetime.datetime): The time when the job execution was enqueued.
        started_at (datetime.datetime): The time when the job execution started.
        status (str): The status of the job execution. Can be one of: QUEUED, IN_PROGRESS, FAILED, SUCCEEDED, CANCELED, TIMED_OUT, REJECTED, or REMOVED.
        status_details (typing.Dict[str, str]): A collection of name-value pairs that describe the status of the job execution.
        thing_name (str): The name of the thing that is executing the job.
        version_number (int): The version of the job execution. Job execution versions are incremented each time they are updated by a device.

    Attributes:
        execution_number (int): A number that identifies a job execution on a device. It can be used later in commands that return or update job execution information.
        job_document (typing.Dict[str, typing.Any]): The content of the job document.
        job_id (str): The unique identifier you assigned to this job when it was created.
        last_updated_at (datetime.datetime): The time when the job execution started.
        queued_at (datetime.datetime): The time when the job execution was enqueued.
        started_at (datetime.datetime): The time when the job execution started.
        status (str): The status of the job execution. Can be one of: QUEUED, IN_PROGRESS, FAILED, SUCCEEDED, CANCELED, TIMED_OUT, REJECTED, or REMOVED.
        status_details (typing.Dict[str, str]): A collection of name-value pairs that describe the status of the job execution.
        thing_name (str): The name of the thing that is executing the job.
        version_number (int): The version of the job execution. Job execution versions are incremented each time they are updated by a device.
    """

    __slots__ = ['execution_number', 'job_document', 'job_id', 'last_updated_at', 'queued_at', 'started_at', 'status', 'status_details', 'thing_name', 'version_number']

    def __init__(self, *args, **kwargs):
        self.execution_number = kwargs.get('execution_number')
        self.job_document = kwargs.get('job_document')
        self.job_id = kwargs.get('job_id')
        self.last_updated_at = kwargs.get('last_updated_at')
        self.queued_at = kwargs.get('queued_at')
        self.started_at = kwargs.get('started_at')
        self.status = kwargs.get('status')
        self.status_details = kwargs.get('status_details')
        self.thing_name = kwargs.get('thing_name')
        self.version_number = kwargs.get('version_number')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['execution_number', 'job_document', 'job_id', 'last_updated_at', 'queued_at', 'started_at', 'status', 'status_details', 'thing_name', 'version_number'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> JobExecutionData
        new = cls()
        val = payload.get('executionNumber')
        if val is not None:
            new.execution_number = val
        val = payload.get('jobDocument')
        if val is not None:
            new.job_document = val
        val = payload.get('jobId')
        if val is not None:
            new.job_id = val
        val = payload.get('lastUpdatedAt')
        if val is not None:
            new.last_updated_at = datetime.datetime.fromtimestamp(val)
        val = payload.get('queuedAt')
        if val is not None:
            new.queued_at = datetime.datetime.fromtimestamp(val)
        val = payload.get('startedAt')
        if val is not None:
            new.started_at = datetime.datetime.fromtimestamp(val)
        val = payload.get('status')
        if val is not None:
            new.status = val
        val = payload.get('statusDetails')
        if val is not None:
            new.status_details = val
        val = payload.get('thingName')
        if val is not None:
            new.thing_name = val
        val = payload.get('versionNumber')
        if val is not None:
            new.version_number = val
        return new

class JobExecutionState(awsiot.ModeledClass):
    """

    Data about the state of a job execution.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        status (str): The status of the job execution. Can be one of: QUEUED, IN_PROGRESS, FAILED, SUCCEEDED, CANCELED, TIMED_OUT, REJECTED, or REMOVED.
        status_details (typing.Dict[str, str]): A collection of name-value pairs that describe the status of the job execution.
        version_number (int): The version of the job execution. Job execution versions are incremented each time they are updated by a device.

    Attributes:
        status (str): The status of the job execution. Can be one of: QUEUED, IN_PROGRESS, FAILED, SUCCEEDED, CANCELED, TIMED_OUT, REJECTED, or REMOVED.
        status_details (typing.Dict[str, str]): A collection of name-value pairs that describe the status of the job execution.
        version_number (int): The version of the job execution. Job execution versions are incremented each time they are updated by a device.
    """

    __slots__ = ['status', 'status_details', 'version_number']

    def __init__(self, *args, **kwargs):
        self.status = kwargs.get('status')
        self.status_details = kwargs.get('status_details')
        self.version_number = kwargs.get('version_number')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['status', 'status_details', 'version_number'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> JobExecutionState
        new = cls()
        val = payload.get('status')
        if val is not None:
            new.status = val
        val = payload.get('statusDetails')
        if val is not None:
            new.status_details = val
        val = payload.get('versionNumber')
        if val is not None:
            new.version_number = val
        return new

class JobExecutionSummary(awsiot.ModeledClass):
    """

    Contains a subset of information about a job execution.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        execution_number (int): A number that identifies a job execution on a device.
        job_id (str): The unique identifier you assigned to this job when it was created.
        last_updated_at (datetime.datetime): The time when the job execution was last updated.
        queued_at (datetime.datetime): The time when the job execution was enqueued.
        started_at (datetime.datetime): The time when the job execution started.
        version_number (int): The version of the job execution. Job execution versions are incremented each time the AWS IoT Jobs service receives an update from a device.

    Attributes:
        execution_number (int): A number that identifies a job execution on a device.
        job_id (str): The unique identifier you assigned to this job when it was created.
        last_updated_at (datetime.datetime): The time when the job execution was last updated.
        queued_at (datetime.datetime): The time when the job execution was enqueued.
        started_at (datetime.datetime): The time when the job execution started.
        version_number (int): The version of the job execution. Job execution versions are incremented each time the AWS IoT Jobs service receives an update from a device.
    """

    __slots__ = ['execution_number', 'job_id', 'last_updated_at', 'queued_at', 'started_at', 'version_number']

    def __init__(self, *args, **kwargs):
        self.execution_number = kwargs.get('execution_number')
        self.job_id = kwargs.get('job_id')
        self.last_updated_at = kwargs.get('last_updated_at')
        self.queued_at = kwargs.get('queued_at')
        self.started_at = kwargs.get('started_at')
        self.version_number = kwargs.get('version_number')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['execution_number', 'job_id', 'last_updated_at', 'queued_at', 'started_at', 'version_number'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> JobExecutionSummary
        new = cls()
        val = payload.get('executionNumber')
        if val is not None:
            new.execution_number = val
        val = payload.get('jobId')
        if val is not None:
            new.job_id = val
        val = payload.get('lastUpdatedAt')
        if val is not None:
            new.last_updated_at = datetime.datetime.fromtimestamp(val)
        val = payload.get('queuedAt')
        if val is not None:
            new.queued_at = datetime.datetime.fromtimestamp(val)
        val = payload.get('startedAt')
        if val is not None:
            new.started_at = datetime.datetime.fromtimestamp(val)
        val = payload.get('versionNumber')
        if val is not None:
            new.version_number = val
        return new

class JobExecutionsChangedEvent(awsiot.ModeledClass):
    """

    Sent whenever a job execution is added to or removed from the list of pending job executions for a thing.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        jobs (typing.Dict[str, typing.List[JobExecutionSummary]]): Map from JobStatus to a list of Jobs transitioning to that status.
        timestamp (datetime.datetime): The time when the message was sent.

    Attributes:
        jobs (typing.Dict[str, typing.List[JobExecutionSummary]]): Map from JobStatus to a list of Jobs transitioning to that status.
        timestamp (datetime.datetime): The time when the message was sent.
    """

    __slots__ = ['jobs', 'timestamp']

    def __init__(self, *args, **kwargs):
        self.jobs = kwargs.get('jobs')
        self.timestamp = kwargs.get('timestamp')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['jobs', 'timestamp'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> JobExecutionsChangedEvent
        new = cls()
        val = payload.get('jobs')
        if val is not None:
            new.jobs = {k: [JobExecutionSummary.from_payload(i) for i in v] for k,v in val.items()}
        val = payload.get('timestamp')
        if val is not None:
            new.timestamp = datetime.datetime.fromtimestamp(val)
        return new

class JobExecutionsChangedSubscriptionRequest(awsiot.ModeledClass):
    """

    Data needed to subscribe to JobExecutionsChanged events.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        thing_name (str): Name of the IoT Thing that you want to subscribe to JobExecutionsChanged events for.

    Attributes:
        thing_name (str): Name of the IoT Thing that you want to subscribe to JobExecutionsChanged events for.
    """

    __slots__ = ['thing_name']

    def __init__(self, *args, **kwargs):
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['thing_name'], args):
            setattr(self, key, val)

class NextJobExecutionChangedEvent(awsiot.ModeledClass):
    """

    Sent whenever there is a change to which job execution is next on the list of pending job executions for a thing, as defined for DescribeJobExecution with jobId $next. This message is not sent when the next job's execution details change, only when the next job that would be returned by DescribeJobExecution with jobId $next has changed.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        execution (JobExecutionData): Contains data about a job execution.
        timestamp (datetime.datetime): The time when the message was sent.

    Attributes:
        execution (JobExecutionData): Contains data about a job execution.
        timestamp (datetime.datetime): The time when the message was sent.
    """

    __slots__ = ['execution', 'timestamp']

    def __init__(self, *args, **kwargs):
        self.execution = kwargs.get('execution')
        self.timestamp = kwargs.get('timestamp')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['execution', 'timestamp'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> NextJobExecutionChangedEvent
        new = cls()
        val = payload.get('execution')
        if val is not None:
            new.execution = JobExecutionData.from_payload(val)
        val = payload.get('timestamp')
        if val is not None:
            new.timestamp = datetime.datetime.fromtimestamp(val)
        return new

class NextJobExecutionChangedSubscriptionRequest(awsiot.ModeledClass):
    """

    Data needed to subscribe to NextJobExecutionChanged events.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        thing_name (str): Name of the IoT Thing that you want to subscribe to NextJobExecutionChanged events for.

    Attributes:
        thing_name (str): Name of the IoT Thing that you want to subscribe to NextJobExecutionChanged events for.
    """

    __slots__ = ['thing_name']

    def __init__(self, *args, **kwargs):
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['thing_name'], args):
            setattr(self, key, val)

class RejectedError(awsiot.ModeledClass):
    """

    Response document containing details about a failed request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): Opaque token that can correlate this response to the original request.
        code (str): Indicates the type of error.
        execution_state (JobExecutionState): A JobExecutionState object. This field is included only when the code field has the value InvalidStateTransition or VersionMismatch.
        message (str): A text message that provides additional information.
        timestamp (datetime.datetime): The date and time the response was generated by AWS IoT.

    Attributes:
        client_token (str): Opaque token that can correlate this response to the original request.
        code (str): Indicates the type of error.
        execution_state (JobExecutionState): A JobExecutionState object. This field is included only when the code field has the value InvalidStateTransition or VersionMismatch.
        message (str): A text message that provides additional information.
        timestamp (datetime.datetime): The date and time the response was generated by AWS IoT.
    """

    __slots__ = ['client_token', 'code', 'execution_state', 'message', 'timestamp']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.code = kwargs.get('code')
        self.execution_state = kwargs.get('execution_state')
        self.message = kwargs.get('message')
        self.timestamp = kwargs.get('timestamp')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'code', 'execution_state', 'message', 'timestamp'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> RejectedError
        new = cls()
        val = payload.get('clientToken')
        if val is not None:
            new.client_token = val
        val = payload.get('code')
        if val is not None:
            new.code = val
        val = payload.get('executionState')
        if val is not None:
            new.execution_state = JobExecutionState.from_payload(val)
        val = payload.get('message')
        if val is not None:
            new.message = val
        val = payload.get('timestamp')
        if val is not None:
            new.timestamp = datetime.datetime.fromtimestamp(val)
        return new

class StartNextJobExecutionResponse(awsiot.ModeledClass):
    """

    Response payload to a StartNextJobExecution request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): A client token used to correlate requests and responses.
        execution (JobExecutionData): Contains data about a job execution.
        timestamp (datetime.datetime): The time when the message was sent to the device.

    Attributes:
        client_token (str): A client token used to correlate requests and responses.
        execution (JobExecutionData): Contains data about a job execution.
        timestamp (datetime.datetime): The time when the message was sent to the device.
    """

    __slots__ = ['client_token', 'execution', 'timestamp']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.execution = kwargs.get('execution')
        self.timestamp = kwargs.get('timestamp')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'execution', 'timestamp'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> StartNextJobExecutionResponse
        new = cls()
        val = payload.get('clientToken')
        if val is not None:
            new.client_token = val
        val = payload.get('execution')
        if val is not None:
            new.execution = JobExecutionData.from_payload(val)
        val = payload.get('timestamp')
        if val is not None:
            new.timestamp = datetime.datetime.fromtimestamp(val)
        return new

class StartNextPendingJobExecutionRequest(awsiot.ModeledClass):
    """

    Data needed to make a StartNextPendingJobExecution request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): Optional. A client token used to correlate requests and responses. Enter an arbitrary value here and it is reflected in the response.
        status_details (typing.Dict[str, str]): A collection of name-value pairs that describe the status of the job execution. If not specified, the statusDetails are unchanged.
        step_timeout_in_minutes (int): Specifies the amount of time this device has to finish execution of this job.
        thing_name (str): IoT Thing the request is relative to.

    Attributes:
        client_token (str): Optional. A client token used to correlate requests and responses. Enter an arbitrary value here and it is reflected in the response.
        status_details (typing.Dict[str, str]): A collection of name-value pairs that describe the status of the job execution. If not specified, the statusDetails are unchanged.
        step_timeout_in_minutes (int): Specifies the amount of time this device has to finish execution of this job.
        thing_name (str): IoT Thing the request is relative to.
    """

    __slots__ = ['client_token', 'status_details', 'step_timeout_in_minutes', 'thing_name']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.status_details = kwargs.get('status_details')
        self.step_timeout_in_minutes = kwargs.get('step_timeout_in_minutes')
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'status_details', 'step_timeout_in_minutes', 'thing_name'], args):
            setattr(self, key, val)

    def to_payload(self):
        # type: () -> typing.Dict[str, typing.Any]
        payload = {} # type: typing.Dict[str, typing.Any]
        if self.client_token is not None:
            payload['clientToken'] = self.client_token
        if self.status_details is not None:
            payload['statusDetails'] = self.status_details
        if self.step_timeout_in_minutes is not None:
            payload['stepTimeoutInMinutes'] = self.step_timeout_in_minutes
        return payload

class StartNextPendingJobExecutionSubscriptionRequest(awsiot.ModeledClass):
    """

    Data needed to subscribe to StartNextPendingJobExecution responses.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        thing_name (str): Name of the IoT Thing that you want to subscribe to StartNextPendingJobExecution response events for.

    Attributes:
        thing_name (str): Name of the IoT Thing that you want to subscribe to StartNextPendingJobExecution response events for.
    """

    __slots__ = ['thing_name']

    def __init__(self, *args, **kwargs):
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['thing_name'], args):
            setattr(self, key, val)

class UpdateJobExecutionRequest(awsiot.ModeledClass):
    """

    Data needed to make an UpdateJobExecution request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): A client token used to correlate requests and responses. Enter an arbitrary value here and it is reflected in the response.
        execution_number (int): Optional. A number that identifies a job execution on a device. If not specified, the latest job execution is used.
        expected_version (int): The expected current version of the job execution. Each time you update the job execution, its version is incremented. If the version of the job execution stored in the AWS IoT Jobs service does not match, the update is rejected with a VersionMismatch error, and an ErrorResponse that contains the current job execution status data is returned.
        include_job_document (bool): Optional. When included and set to true, the response contains the JobDocument. The default is false.
        include_job_execution_state (bool): Optional. When included and set to true, the response contains the JobExecutionState field. The default is false.
        job_id (str): The unique identifier assigned to this job when it was created.
        status (str): The new status for the job execution (IN_PROGRESS, FAILED, SUCCEEDED, or REJECTED). This must be specified on every update.
        status_details (typing.Dict[str, str]): A collection of name-value pairs that describe the status of the job execution. If not specified, the statusDetails are unchanged.
        step_timeout_in_minutes (int): Specifies the amount of time this device has to finish execution of this job. If the job execution status is not set to a terminal state before this timer expires, or before the timer is reset (by again calling UpdateJobExecution, setting the status to IN_PROGRESS and specifying a new timeout value in this field) the job execution status is set to TIMED_OUT. Setting or resetting this timeout has no effect on the job execution timeout that might have been specified when the job was created (by using CreateJob with the timeoutConfig).
        thing_name (str): The name of the thing associated with the device.

    Attributes:
        client_token (str): A client token used to correlate requests and responses. Enter an arbitrary value here and it is reflected in the response.
        execution_number (int): Optional. A number that identifies a job execution on a device. If not specified, the latest job execution is used.
        expected_version (int): The expected current version of the job execution. Each time you update the job execution, its version is incremented. If the version of the job execution stored in the AWS IoT Jobs service does not match, the update is rejected with a VersionMismatch error, and an ErrorResponse that contains the current job execution status data is returned.
        include_job_document (bool): Optional. When included and set to true, the response contains the JobDocument. The default is false.
        include_job_execution_state (bool): Optional. When included and set to true, the response contains the JobExecutionState field. The default is false.
        job_id (str): The unique identifier assigned to this job when it was created.
        status (str): The new status for the job execution (IN_PROGRESS, FAILED, SUCCEEDED, or REJECTED). This must be specified on every update.
        status_details (typing.Dict[str, str]): A collection of name-value pairs that describe the status of the job execution. If not specified, the statusDetails are unchanged.
        step_timeout_in_minutes (int): Specifies the amount of time this device has to finish execution of this job. If the job execution status is not set to a terminal state before this timer expires, or before the timer is reset (by again calling UpdateJobExecution, setting the status to IN_PROGRESS and specifying a new timeout value in this field) the job execution status is set to TIMED_OUT. Setting or resetting this timeout has no effect on the job execution timeout that might have been specified when the job was created (by using CreateJob with the timeoutConfig).
        thing_name (str): The name of the thing associated with the device.
    """

    __slots__ = ['client_token', 'execution_number', 'expected_version', 'include_job_document', 'include_job_execution_state', 'job_id', 'status', 'status_details', 'step_timeout_in_minutes', 'thing_name']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.execution_number = kwargs.get('execution_number')
        self.expected_version = kwargs.get('expected_version')
        self.include_job_document = kwargs.get('include_job_document')
        self.include_job_execution_state = kwargs.get('include_job_execution_state')
        self.job_id = kwargs.get('job_id')
        self.status = kwargs.get('status')
        self.status_details = kwargs.get('status_details')
        self.step_timeout_in_minutes = kwargs.get('step_timeout_in_minutes')
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'execution_number', 'expected_version', 'include_job_document', 'include_job_execution_state', 'job_id', 'status', 'status_details', 'step_timeout_in_minutes', 'thing_name'], args):
            setattr(self, key, val)

    def to_payload(self):
        # type: () -> typing.Dict[str, typing.Any]
        payload = {} # type: typing.Dict[str, typing.Any]
        if self.client_token is not None:
            payload['clientToken'] = self.client_token
        if self.execution_number is not None:
            payload['executionNumber'] = self.execution_number
        if self.expected_version is not None:
            payload['expectedVersion'] = self.expected_version
        if self.include_job_document is not None:
            payload['includeJobDocument'] = self.include_job_document
        if self.include_job_execution_state is not None:
            payload['includeJobExecutionState'] = self.include_job_execution_state
        if self.status is not None:
            payload['status'] = self.status
        if self.status_details is not None:
            payload['statusDetails'] = self.status_details
        if self.step_timeout_in_minutes is not None:
            payload['stepTimeoutInMinutes'] = self.step_timeout_in_minutes
        return payload

class UpdateJobExecutionResponse(awsiot.ModeledClass):
    """

    Response payload to an UpdateJobExecution request.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        client_token (str): A client token used to correlate requests and responses.
        execution_state (JobExecutionState): Contains data about the state of a job execution.
        job_document (typing.Dict[str, typing.Any]): A UTF-8 encoded JSON document that contains information that your devices need to perform the job.
        timestamp (datetime.datetime): The time when the message was sent.

    Attributes:
        client_token (str): A client token used to correlate requests and responses.
        execution_state (JobExecutionState): Contains data about the state of a job execution.
        job_document (typing.Dict[str, typing.Any]): A UTF-8 encoded JSON document that contains information that your devices need to perform the job.
        timestamp (datetime.datetime): The time when the message was sent.
    """

    __slots__ = ['client_token', 'execution_state', 'job_document', 'timestamp']

    def __init__(self, *args, **kwargs):
        self.client_token = kwargs.get('client_token')
        self.execution_state = kwargs.get('execution_state')
        self.job_document = kwargs.get('job_document')
        self.timestamp = kwargs.get('timestamp')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['client_token', 'execution_state', 'job_document', 'timestamp'], args):
            setattr(self, key, val)

    @classmethod
    def from_payload(cls, payload):
        # type: (typing.Dict[str, typing.Any]) -> UpdateJobExecutionResponse
        new = cls()
        val = payload.get('clientToken')
        if val is not None:
            new.client_token = val
        val = payload.get('executionState')
        if val is not None:
            new.execution_state = JobExecutionState.from_payload(val)
        val = payload.get('jobDocument')
        if val is not None:
            new.job_document = val
        val = payload.get('timestamp')
        if val is not None:
            new.timestamp = datetime.datetime.fromtimestamp(val)
        return new

class UpdateJobExecutionSubscriptionRequest(awsiot.ModeledClass):
    """

    Data needed to subscribe to UpdateJobExecution responses.

    All attributes are None by default, and may be set by keyword in the constructor.

    Keyword Args:
        job_id (str): Job ID that you want to subscribe to UpdateJobExecution response events for.
        thing_name (str): Name of the IoT Thing that you want to subscribe to UpdateJobExecution response events for.

    Attributes:
        job_id (str): Job ID that you want to subscribe to UpdateJobExecution response events for.
        thing_name (str): Name of the IoT Thing that you want to subscribe to UpdateJobExecution response events for.
    """

    __slots__ = ['job_id', 'thing_name']

    def __init__(self, *args, **kwargs):
        self.job_id = kwargs.get('job_id')
        self.thing_name = kwargs.get('thing_name')

        # for backwards compatibility, read any arguments that used to be accepted by position
        for key, val in zip(['job_id', 'thing_name'], args):
            setattr(self, key, val)

class JobStatus:
    """

    The status of the job execution.

    """

    CANCELED = 'CANCELED'
    """
    """

    FAILED = 'FAILED'
    """
    """

    QUEUED = 'QUEUED'
    """
    """

    IN_PROGRESS = 'IN_PROGRESS'
    """
    """

    SUCCEEDED = 'SUCCEEDED'
    """
    """

    TIMED_OUT = 'TIMED_OUT'
    """
    """

    REJECTED = 'REJECTED'
    """
    """

    REMOVED = 'REMOVED'
    """
    """

class RejectedErrorCode:
    """

    A value indicating the kind of error encountered while processing an AWS IoT Jobs request

    """

    INTERNAL_ERROR = 'InternalError'
    """
    There was an internal error during the processing of the request.
    """

    INVALID_JSON = 'InvalidJson'
    """
    The contents of the request could not be interpreted as valid UTF-8-encoded JSON.
    """

    INVALID_REQUEST = 'InvalidRequest'
    """
    The contents of the request were invalid. The message contains details about the error.
    """

    INVALID_STATE_TRANSITION = 'InvalidStateTransition'
    """
    An update attempted to change the job execution to a state that is invalid because of the job execution's current state. In this case, the body of the error message also contains the executionState field.
    """

    RESOURCE_NOT_FOUND = 'ResourceNotFound'
    """
    The JobExecution specified by the request topic does not exist.
    """

    VERSION_MISMATCH = 'VersionMismatch'
    """
    The expected version specified in the request does not match the version of the job execution in the AWS IoT Jobs service. In this case, the body of the error message also contains the executionState field.
    """

    INVALID_TOPIC = 'InvalidTopic'
    """
    The request was sent to a topic in the AWS IoT Jobs namespace that does not map to any API.
    """

    REQUEST_THROTTLED = 'RequestThrottled'
    """
    The request was throttled.
    """

    TERMINAL_STATE_REACHED = 'TerminalStateReached'
    """
    Occurs when a command to describe a job is performed on a job that is in a terminal state.
    """




"""
Classes for building a service that uses the event-stream RPC protocol.
"""
# Copyright Amazon.com, Inc. or its affiliates. All Rights Reserved.
# SPDX-License-Identifier: Apache-2.0.

from awscrt.eventstream import Header, HeaderType
import awscrt.eventstream.rpc as protocol
from awscrt.io import (ClientBootstrap, SocketOptions, TlsConnectionOptions)
from concurrent.futures import Future
from enum import Enum
import json
import logging
from threading import Lock
from typing import (Any, Callable, Dict, Optional, Sequence)

VERSION_TUPLE = (0, 1, 0)
VERSION_STRING = "{v[0]}.{v[1]}.{v[2]}".format(v=VERSION_TUPLE)

VERSION_HEADER = ":version"
CONTENT_TYPE_HEADER = ":content-type"
CONTENT_TYPE_APPLICATION_TEXT = "text/plain"
CONTENT_TYPE_APPLICATION_JSON = "application/json"
SERVICE_MODEL_TYPE_HEADER = "service-model-type"


logger = logging.getLogger(__name__)


class ConnectionClosedError(RuntimeError):
    """
    Connection is closed
    """
    pass


class StreamClosedError(RuntimeError):
    """
    Stream is closed
    """
    pass


class EventStreamError(RuntimeError):
    """
    For connection-level errors.
    """
    pass


class EventStreamOperationError(RuntimeError):
    """
    Base for all errors that come across the wire.

    These are not necessarily modeled shapes.
    """
    pass


class AccessDeniedError(EventStreamOperationError):
    """
    Access Denied
    """

    def __init__(self, *args):
        super().__init__('aws#AccessDenied', *args)


class UnmappedDataError(RuntimeError):
    """
    Received data that does not map to a known model type.
    """
    pass


class SerializeError(RuntimeError):
    """
    Error serializing data to send.
    """
    pass


class DeserializeError(RuntimeError):
    """
    Error deserializing received data.
    """
    pass


class LifecycleHandler:
    """
    Base class for handling connection events.

    Inherit from this class and override methods to handle connection events.
    All callbacks for this connection will be invoked on the same thread.
    If the connection attempt fails, no callbacks will be invoked.
    If the connection attempt succeeds, :meth:`on_connect()` will be the first callback
    invoked and :meth:`on_disconnect()` will always be the last.

    Note that an open network connection MUST be closed via :meth:`Connection.close()`
    to avoid leaking resources.
    """

    def on_connect(self):
        """
        Invoked when the connection has been fully established.

        This will always be the first callback invoked on the handler.
        This will not be invoked if the connection attempt failed.
        """
        pass

    def on_disconnect(self, reason: Optional[Exception]):
        """
        Invoked when an open connection has disconnected.

        This will always be the last callback invoked on the handler.
        This will not be invoked if the connection attempt failed.

        Args:
            reason: Reason will be `None` if the user initiated the shutdown,
                otherwise the reason will be an Exception.
        """
        pass

    def on_error(self, error: Exception) -> bool:
        """
        Invoked when a connection-level error occurs.

        Args:
            reason: An Exception explaining the error

        Returns:
            True if the connection should be terminated as a result of the error,
            or False if the connection should continute.
        """
        pass

    def on_ping(self, headers: Sequence[Header], payload: bytes):
        """
        Invoked when a ping is received.
        """
        pass


class MessageAmendment:
    """
    Data to add to an event-stream message.

    Args:
        headers: Headers to add (optional)

        payload: Binary payload data (optional)
    """

    def __init__(self, *, headers: Optional[Sequence[Header]] = None, payload: Optional[bytes] = None):
        #: Headers to add
        self.headers = headers # type: Optional[Sequence[Header]]

        #: Binary payload data
        self.payload = payload # type: Optional[bytes]

    @staticmethod
    def create_static_authtoken_amender(authtoken: str) -> Callable[[], 'MessageAmendment']:
        """
        Create function that amends payload: `b'{"authToken": "..."}'`

        Args:
            authtoken: value of "authToken" in the payload.
                The same value is always used, even if the amender
                is called multiple times over the life of the application.

        Returns:
            The result is appropriate for passing to the :class:`Connection`'s
            `connect_message_amender` init arg.
        """
        def _amend():
            payload_str = '{"authToken": "%s"}' % authtoken
            return MessageAmendment(payload=payload_str.encode())

        return _amend


class _ClientState(Enum):
    DISCONNECTED = 1
    CONNECTING_TO_SOCKET = 2
    WAITING_FOR_CONNECT_ACK = 3
    CONNECTED = 4
    DISCONNECTING = 5


class _ProtocolConnectionHandler(protocol.ClientConnectionHandler):
    def __init__(self, owner, lifecycle_handler):
        self.owner = owner
        self.lifecycle_handler = lifecycle_handler

    def on_connection_setup(self, connection, error, **kwargs):
        # if error is set, socket connection failed
        if error:
            logger.error("%r failed to establish connection: %r", self.owner, error)
            synced = self.owner._synced
            with synced.lock:
                connect_future = synced.connect_future

                synced.connect_future = None
                synced.current_handler = None
                synced.close_reason = None
                synced.closed_future.set_exception(error)
                synced.state = _ClientState.DISCONNECTED
            # complete future after lock is released
            connect_future.set_exception(error)
            return

        # error is not set, so socket connection is established.
        # next step is to send CONNECT message
        try:
            logger.debug("%r connection established, sending CONNECT message", self.owner)
            # invoke callback outside of holding the lock
            if self.owner._connect_message_amender:
                amendment = self.owner._connect_message_amender()
            else:
                amendment = MessageAmendment()

            with self.owner._synced as synced:
                synced.current_connection = connection
                # check if close() was called before connection established
                if synced.state == _ClientState.DISCONNECTING:
                    logger.debug("%r close() has been called, shutting down", self.owner)
                    connection.close()
                else:
                    headers = [Header.from_string(
                        VERSION_HEADER, VERSION_STRING)]
                    # don't allow amendment to override required headers
                    existing_names = [header.name.lower() for header in headers]
                    if amendment.headers:
                        for header in amendment.headers:
                            if header.name.lower() not in existing_names:
                                headers.append(header)

                    connection.send_protocol_message(
                        headers=headers, payload=amendment.payload,
                        message_type=protocol.MessageType.CONNECT)
                    synced.state = _ClientState.WAITING_FOR_CONNECT_ACK
        except Exception as e:
            logger.debug("%r failure attempting to send CONNECT: %r", self.owner, e)
            with self.owner._synced as synced:
                synced.state = _ClientState.DISCONNECTING
                synced.current_connection = connection
                synced.close_reason = e
                connection.close()

    def on_connection_shutdown(self, reason, **kwargs):
        connect_future = None
        with self.owner._synced as synced:
            connect_future = synced.connect_future
            if reason is None:
                reason = synced.close_reason

            synced.connect_future = None
            synced.state = _ClientState.DISCONNECTED
            synced.current_handler = None
            synced.current_connection = None
            synced.close_reason = None
            if reason is None:
                synced.closed_future.set_result(None)
            else:
                synced.closed_future.set_exception(reason)

        # complete futures, and invoke callbacks, after lock is released

        # if connect_future still exists, mark that setup was a failure
        if connect_future:
            # if user called close() without a reason,
            # set a reason that the setup_future has failed
            if reason is None:
                reason = ConnectionClosedError("close() called during connection setup")
            logger.error("%r connect failed: %r", self.owner, reason)
            connect_future.set_exception(reason)
        else:
            # connect_future no longer exists, which means on_setup fired,
            # which means on_disconnect should fire now
            if reason:
                logger.info("%r disconnected, reason: %r", self.owner, reason)
            else:
                logger.info("%r disconnected", self.owner)
            self.lifecycle_handler.on_disconnect(reason)

    def on_protocol_message(self, headers, payload, message_type, flags, **kwargs):
        try:
            logger.debug("%r received %s headers=%s", self.owner, message_type.name, headers)

            # protocol enforces that CONNECT_ACK is first msg received
            if message_type == protocol.MessageType.CONNECT_ACK:
                connect_future = None
                with self.owner._synced as synced:
                    if synced.state == _ClientState.WAITING_FOR_CONNECT_ACK:
                        if (flags & protocol.MessageFlag.CONNECTION_ACCEPTED):
                            connect_future = synced.connect_future
                            synced.connect_future = None
                            synced.state = _ClientState.CONNECTED
                        else:
                            synced.state = _ClientState.DISCONNECTING
                            synced.close_reason = AccessDeniedError(
                                "Connection access denied to event stream RPC server")
                            synced.current_connection.close()
                # complete future and invoke callback after lock is released
                if connect_future:
                    logger.info("%r connected", self.owner)
                    connect_future.set_result(None)
                    self.lifecycle_handler.on_connect()
            elif message_type == protocol.MessageType.PING_RESPONSE:
                pass
            elif message_type == protocol.MessageType.PING:
                self.lifecycle_handler.on_ping(headers, payload)
            elif message_type in (protocol.MessageType.PROTOCOL_ERROR, protocol.MessageType.INTERNAL_ERROR):
                error = EventStreamError(message_type, headers, payload)
                # If callback returns True (or forgets to return a value)
                # then close connection due to error
                return_val = self.lifecycle_handler.on_error(error)
                if return_val or return_val is None:
                    self.owner.close(error)
        except Exception as e:
            logger.error("%r closing due to exception from LifecycleHandler callback: %r", self.owner, e)
            self.owner.close(e)


class Connection:
    """
    A client connection to event-stream RPC service.

    connect() must be called to open the network connection before interacting
    with the service.

    Note that close() MUST be called to end an open network connection.
    Failure to do so will result in leaked resources.

    Reconnect is possible by calling connect() again after the connection
    has finished closing/disconnecting.

    Args:
        host_name: Remote host name.

        port: Remote port.

        bootstrap: ClientBootstrap to use when initiating socket connection.

        socket_options: Optional socket options.
            If None is provided, the default options are used.

        tls_connection_options: Optional TLS connection options.
            If None is provided, then the connection will be attempted over
            plain-text.

        connect_message_amender: Optional callable that should return a
            :class:`MessageAmendment` for the
            :attr:`~awscrt.eventstream.rpc.MessageType.CONNECT` message.
            This callable will be invoked whenever a network connection is
            being established.
    """

    class _Synced:
        """
        Helper class holds all values that must not be read/written without a lock.
        """

        def __init__(self):
            self.lock = Lock()
            self.state = _ClientState.DISCONNECTED
            self.current_handler = None
            self.current_connection = None
            self.connect_future = None
            self.close_reason = None
            self.closed_future = Future()
            # closed future starts out as complete,
            # since we haven't even tried to connect yet
            self.closed_future.set_result(None)

        def __enter__(self):
            self.lock.acquire()
            return self

        def __exit__(self, type, value, tb):
            self.lock.release()

    def __init__(self,
                 *,
                 host_name: str,
                 port: int,
                 bootstrap: ClientBootstrap,
                 socket_options: Optional[SocketOptions] = None,
                 tls_connection_options: Optional[TlsConnectionOptions] = None,
                 connect_message_amender: Optional[Callable[[], MessageAmendment]] = None):

        self.host_name = host_name
        self.port = port
        self._bootstrap = bootstrap
        self._socket_options = socket_options
        self._tls_connection_options = tls_connection_options
        self._connect_message_amender = connect_message_amender

        self._synced = Connection._Synced()

    def connect(self, lifecycle_handler: LifecycleHandler) -> Future:
        """
        Asynchronously open a network connection.

        Note that close() MUST be called to end a network connection
        that is open (or in the process of connecting).
        Failure to do so will result in leaked resources.

        Args:
            lifecycle_handler: Handler for events over the course of this
                network connection. See :class:`LifecycleHandler` for more info.
                Handler methods will only be invoked if the connect attempt
                succeeds.

        Returns:
            A Future which completes when the connection succeeds or fails.
            If successful, the Future will contain None.
            Otherwise it will contain an exception explaining the reason
            for failure.
        """
        future = Future()
        future.set_running_or_notify_cancel()  # prevent cancellation
        with self._synced as synced:
            old_closed_future = synced.closed_future
            if synced.state != _ClientState.DISCONNECTED:
                raise RuntimeError("Connection already in progress")
            try:
                synced.current_handler = _ProtocolConnectionHandler(
                    self, lifecycle_handler)
                synced.connect_future = future
                synced.state = _ClientState.CONNECTING_TO_SOCKET
                # start new closed_future
                synced.closed_future = Future()
                synced.closed_future.set_running_or_notify_cancel()
                protocol.ClientConnection.connect(
                    handler=synced.current_handler,
                    host_name=self.host_name,
                    port=self.port,
                    bootstrap=self._bootstrap,
                    socket_options=self._socket_options,
                    tls_connection_options=self._tls_connection_options)
            except Exception as e:
                synced.current_handler = None
                synced.connect_future = None
                synced.closed_future = old_closed_future
                synced.state = _ClientState.DISCONNECTED
                raise e
        return future

    def close(self, reason: Optional[Exception] = None) -> Future:
        """
        Close the connection.

        Shutdown is asynchronous. This call has no effect if the connection
        is already closed or closing.

        Args:
            reason: If set, the connection will
                close with this error as the reason (unless
                it was already closing for another reason).

        Returns:
            The future which will complete
            when the shutdown process is done. The future will have an
            exception if shutdown was caused by an error, or a result
            of None if the shutdown was clean and user-initiated.
        """
        with self._synced as synced:
            if synced.state == _ClientState.DISCONNECTED:
                # do nothing, already disconnected
                pass
            elif synced.state == _ClientState.DISCONNECTING:
                # do nothing, already disconnecting for some other reason
                pass
            else:
                synced.close_reason = reason
                synced.state = _ClientState.DISCONNECTING
                # close connection if it exists.
                # if it doesn't exist yet, then it's connecting right now
                # and will get closed the moment it exists.
                if synced.current_connection:
                    synced.current_connection.close()

            return synced.closed_future

    def _send_protocol_message(self, data, message_type):
        with self._synced as synced:
            if synced.state != _ClientState.CONNECTED:
                raise ConnectionClosedError()
            return synced.current_connection.send_protocol_message(
                headers=data.headers if data else None,
                payload=data.payload if data else None,
                message_type=message_type)

    def send_ping(self, data: Optional[MessageAmendment] = None) -> Future:
        return self._send_protocol_message(data, protocol.MessageType.PING)

    def send_ping_response(self, data: Optional[MessageAmendment] = None) -> Future:
        return self._send_protocol_message(data, protocol.MessageType.PING_RESPONSE)

    def _new_stream(self, handler: protocol.ClientContinuationHandler) -> protocol.ClientContinuation:
        # public or private?
        with self._synced as synced:
            if synced.state != _ClientState.CONNECTED:
                raise ConnectionClosedError()
            return synced.current_connection.new_stream(handler)

    def __repr__(self):
        return "<%s at %#x %s:%d>" % (self.__class__.__name__, id(self), self.host_name, self.port)


class Shape:
    """
    Base class for shapes serialized by a service
    """
    @classmethod
    def _model_name(cls):
        raise NotImplementedError(cls.__name__ + " must override _model_name()")

    @classmethod
    def _from_payload(cls, payload):
        raise NotImplementedError(cls.__name__ + " must override _from_payload()")

    def _to_payload(self):
        raise NotImplementedError(self.__class__.__name__ + " must override _to_payload()")


class ErrorShape(Shape, EventStreamOperationError):
    """
    Base class for all error shapes serialized by a service
    """
    pass


class ShapeIndex:
    """
    Catalog of all shapes serialized by this service
    """

    def __init__(self, shape_types: Sequence[type]):
        self._shapes_type_by_name = {i._model_name(): i for i in shape_types}

    def find_shape_type(self, model_name: str) -> type:
        """
        Returns Shape type with given model_name, or None
        """
        return self._shapes_type_by_name.get(model_name)


class StreamResponseHandler:
    """
    Base class for all operation stream handlers.

    For operations with a streaming response (0+ messages that may arrive
    after the initial response).
    """

    def on_stream_event(self, event: Shape) -> None:
        pass

    def on_stream_error(self, error: Exception) -> bool:
        return True

    def on_stream_closed(self) -> None:
        pass


class Operation:
    """
    Base class for an operation.
    """
    @classmethod
    def _model_name(cls) -> str:
        raise NotImplementedError(cls.__name__ + " must override _model_name()")

    @classmethod
    def _request_type(cls) -> type:
        raise NotImplementedError(cls.__name__ + " must override _request_type()")

    @classmethod
    def _request_stream_type(cls) -> type:
        return None

    @classmethod
    def _response_type(cls) -> type:
        raise NotImplementedError(cls.__name__ + " must override _response_type()")

    @classmethod
    def _response_stream_type(cls) -> type:
        return None


class ClientOperation(Operation):
    """
    Base class for a client operation.

    Nearly all functions are private/protected. Child classes should
    rewrite public API to properly document the types they deal with.
    """

    def __init__(self, stream_handler: StreamResponseHandler, shape_index: ShapeIndex, connection: Connection):
        # do not instantiate directly, created by ServiceClient.new_operation()
        # all callbacks that modify state fire on the same thread,
        # so don't need locks to protect members
        self._stream_handler = stream_handler
        self._shape_index = shape_index
        self._message_count = 0
        self._closed_future = Future()
        self._closed_future.set_running_or_notify_cancel()  # prevent cancel
        self._initial_response_future = Future()
        self._initial_response_future.set_running_or_notify_cancel()  # prevent cancel
        protocol_handler = _ProtocolContinuationHandler(self)
        self._continuation = connection._new_stream(protocol_handler)

    def _activate(self, request: Shape) -> Future:
        headers = [Header.from_string(CONTENT_TYPE_HEADER,
                                      CONTENT_TYPE_APPLICATION_JSON),
                   Header.from_string(SERVICE_MODEL_TYPE_HEADER,
                                      request._model_name())]
        payload = self._json_payload_from_shape(request)
        logger.debug("%r sending request APPLICATION_MESSAGE %s %r", self, headers, payload)
        return self._continuation.activate(
            operation=self._model_name(),
            headers=headers,
            payload=payload,
            message_type=protocol.MessageType.APPLICATION_MESSAGE)

    def _send_stream_event(self, event: Shape) -> Future:
        headers = [Header.from_string(CONTENT_TYPE_HEADER,
                                      CONTENT_TYPE_APPLICATION_JSON),
                   Header.from_string(SERVICE_MODEL_TYPE_HEADER,
                                      event._model_name())]
        payload = self._json_payload_from_shape(event)
        logger.debug("%r sending event APPLICATION_MESSAGE %s %r", self, headers, payload)
        return self._continuation.send_message(
            headers=headers,
            payload=payload,
            message_type=protocol.MessageType.APPLICATION_MESSAGE)

    def _get_response(self) -> Future:
        return self._initial_response_future

    def close(self) -> Future:
        try:
            # try to send empty APPLICATION_MESSAGE with TERMINATE_STREAM flag.
            # this fails if stream is already closed, so just ignore errors.
            self._continuation.send_message(
                message_type=protocol.MessageType.APPLICATION_MESSAGE,
                flags=protocol.MessageFlag.TERMINATE_STREAM)
        except Exception:
            pass
        return self._closed_future

    def _find_header(self, headers, name, header_type=HeaderType.STRING):
        """Return header value, or None"""
        name_lower = name.lower()
        for header in headers:
            if header.name.lower() == name_lower:
                if header.type == header_type:
                    return header.value
        return None

    def _shape_from_json_payload(self, payload_bytes, shape_type):
        try:
            payload_str = payload_bytes.decode()
            payload_obj = json.loads(payload_str)
            shape = shape_type._from_payload(payload_obj)
            return shape
        except Exception as e:
            raise DeserializeError("Failed to deserialize %s" % shape_type._model_name(), e, payload_bytes)

    def _json_payload_from_shape(self, shape):
        try:
            payload_obj = shape._to_payload()
            payload_str = json.dumps(payload_obj)
            payload_bytes = payload_str.encode()
            return payload_bytes
        except Exception as e:
            raise SerializeError("Failed to serialize", shape, e)

    def _on_continuation_message(
            self,
            headers: Sequence[Header],
            payload: bytes,
            message_type: protocol.MessageType,
            flags: int,
            **kwargs):
        self._message_count += 1
        logger.debug("%r received #%d %s %s %r", self, self._message_count, message_type.name, headers, payload)
        try:
            model_name = self._find_header(headers, SERVICE_MODEL_TYPE_HEADER)
            if model_name is None:
                if flags & protocol.MessageFlag.TERMINATE_STREAM:
                    # it's ok for a TERMINATE_STREAM message to be empty
                    return
                msg = "Missing header: " + SERVICE_MODEL_TYPE_HEADER
                raise UnmappedDataError(msg, headers, payload)

            content_type = self._find_header(headers, CONTENT_TYPE_HEADER)
            if content_type is None:
                msg = "Missing header: " + CONTENT_TYPE_HEADER
                raise UnmappedDataError(msg, headers, payload)
            if content_type != CONTENT_TYPE_APPLICATION_JSON:
                msg = "Unexpected {}: '{}', expected: '{}'".format(
                    CONTENT_TYPE_HEADER, content_type, CONTENT_TYPE_APPLICATION_JSON)
                raise UnmappedDataError(msg, headers, payload)

            if message_type == protocol.MessageType.APPLICATION_MESSAGE:
                self._handle_data(model_name, payload)
                return

            # otherwise it's an APPLICATION_ERROR
            found_type = self._shape_index.find_shape_type(model_name)
            if found_type is None:
                msg = "Unknown error type: {}".format(model_name)
                raise UnmappedDataError(msg, payload)
            if not issubclass(found_type, Exception):
                msg = "Unexpected type: {} sent as APPLICATION_ERROR, expected subclass of Exception".format(model_name)
                raise UnmappedDataError(msg, payload)
            shape = self._shape_from_json_payload(payload, found_type)
            raise shape
        except Exception as e:
            self._handle_error(e, flags)

    def _handle_data(self, model_name, payload):
        """
        Pass APPLICATION_MESSAGE payload along as a 1st response,
        or subsequent stream-event. Any exceptions raised by this function
        will be passed to _handle_error().
        """
        if self._message_count == 1:
            # 1st message is "response"
            expected_type = self._response_type()
            expected_name = expected_type._model_name()
            if model_name != expected_name:
                msg = "Unexpected response type: {}, expected: {}".format(model_name, expected_name)
                raise UnmappedDataError(msg, payload)
            shape = self._shape_from_json_payload(payload, expected_type)
            self._initial_response_future.set_result(shape)
        else:
            # messages after the 1st are "stream events"
            expected_type = self._response_stream_type()
            if expected_type is None:
                msg = "Operation does not support response stream events, received type: {}".format(model_name)
                raise UnmappedDataError(msg, payload)
            expected_name = expected_type._model_name()
            if model_name != expected_name:
                msg = "Unexpected response stream event type: {}, expected: {}".format(model_name, expected_name)
                raise UnmappedDataError(msg, payload)
            shape = self._shape_from_json_payload(payload, expected_type)
            self._stream_handler.on_stream_event(shape)

    def _handle_error(self, error, message_flags):
        """
        Pass along an APPLICATION_ERROR payload, or an exception encountered while
        processing an APPLICATION_MESSAGE, as a failed 1st response
        or a stream-error.
        """
        stream_already_terminated = message_flags & protocol.MessageFlag.TERMINATE_STREAM
        try:
            if self._message_count == 1:
                # error from 1st message is "response" error.
                self._initial_response_future.set_exception(error)
                # errors on initial response must terminate the stream
                if not stream_already_terminated:
                    self.close()
            elif self._stream_handler is not None:
                # error from subsequent messages are "stream errors"
                # If this callback returns True (or forgets to return a value)
                # then close the stream
                return_val = self._stream_handler.on_stream_error(error)
                if return_val or return_val is None:
                    if not stream_already_terminated:
                        self.close()
            else:
                # this operation did not expect more than 1 message
                raise error
        except Exception:
            logger.exception("%r unhandled exception while receiving message", self)

    def _on_continuation_closed(self, **kwargs) -> None:
        logger.debug("%r closed", self)
        if not self._initial_response_future.done():
            self._initial_response_future.set_exception(StreamClosedError())

        self._closed_future.set_result(None)

        if self._stream_handler:
            try:
                self._stream_handler.on_stream_closed()
            except Exception:
                logger.exception("%r unhandled exception calling callback", self)


class _ProtocolContinuationHandler(protocol.ClientContinuationHandler):
    """Passes raw RPC stream/continuation events along to Operation"""

    def __init__(self, operation: Operation):
        self.operation = operation

    def on_continuation_message(self, *args, **kwargs):
        self.operation._on_continuation_message(*args, **kwargs)

    def on_continuation_closed(self, *args, **kwargs):
        self.operation._on_continuation_closed(*args, **kwargs)
        # break circular reference between: ClientOperation, _ProtocolContinuationHandler, ClientContinuation
        # so that garbage collector can clean them up
        self.operation = None


class Client:
    """
    Base class for a service client.

    Child class should add public API functions for each operation.
    """

    def __init__(self, connection: Connection, shape_index: ShapeIndex):
        self._connection = connection
        self._shape_index = shape_index

    def close(self, reason: Optional[Exception] = None) -> Future:
        """
        Close the connection.

        Shutdown is asynchronous. This call has no effect if the connection
        is already closed or closing.

        Args:
            reason: If set, the connection will
                close with this error as the reason (unless
                it was already closing for another reason).

        Returns:
            The future which will complete
            when the shutdown process is done. The future will have an
            exception if shutdown was caused by an error, or a result
            of None if the shutdown was clean and user-initiated.
        """
        return self._connection.close(reason=reason)

    def _new_operation(self, operation_type: type, stream_handler: StreamResponseHandler = None):
        return operation_type(stream_handler, self._shape_index, self._connection)

    @classmethod
    def _model_name(cls):
        raise NotImplementedError(cls.__name__ + " must override _model_name()")
